from qulacs import QuantumCircuit

from qulacsvis.visualization.circuit_parser import CircuitParser


def _generate_latex_source(circuit: QuantumCircuit) -> str:
    gate_dict = {
        "I": r"{I}",
        "X": r"{X}",
        "Y": r"{Y}",
        "Z": r"{Z}",
        "H": r"{H}",
        "S": r"{S}",
        "Sdag": r"{S^\dag}",
        "T": r"{T}",
        "Tdag": r"{T^\dag}",
        "sqrtX": r"{\sqrt{X}}",
        "sqrtXdag": r"{\sqrt{X^\dag}}",
        "sqrtY": r"{\sqrt{Y}}",
        "sqrtYdag": r"{\sqrt{Y^\dag}}",
        "Projection-0": r"{P0}",
        "Projection-1": r"{P1}",
        "U1": r"{U1}",
        "U2": r"{U2}",
        "U3": r"{U3}",
        "X-rotation": r"{RX}",
        "Y-rotation": r"{RY}",
        "Z-rotation": r"{RZ}",
        "Pauli": r"{Pauli}",
        "Pauli-rotation": r"{PR}",
        "CZ": r"{CZ}",
        "CNOT": r"\targ",
        "SWAP": r"\qswap",
        "Reflection": r"{Ref}",
        "ReversibleBoolean": r"{ReB}",
        "DenseMatrix": r"{DeM}",
        "DinagonalMatrix": r"{DiM}",
        "SparseMatrix": r"{SpM}",
        "Generic gate": r"{GeG}",
        "ParametricRX": r"{pRX}",
        "ParametricRY": r"{pRY}",
        "ParametricRZ": r"{pRZ}",
        "ParametricPauliRotation": r"{pPR}",
        "wire": r"\qw",
        "ghost": "ghost",
    }
    parser = CircuitParser(circuit)
    qubit_count = parser.qubit_count

    gate_latex = [
        [
            r"        \nghost{ {q}_{" + str(i) + r"} : }",
            r"\lstick{ {q}_{" + str(i) + r"} :  }",
        ]
        for i in range(qubit_count)
    ]

    for i in range(len(parser.gate_info[0])):
        gate_latex_part = [r"\qw" for _ in range(qubit_count)]
        for j in range(qubit_count):
            gate_data = parser.gate_info[j][i]
            name_latex_part = gate_dict[gate_data["raw_text"]]
            target_index_list = gate_data["target_bit"]
            control_index_list = gate_data["control_bit"]
            if name_latex_part == r"\qswap":
                for target_index in target_index_list:
                    gate_latex_part[target_index] = name_latex_part
                gate_latex_part[target_index_list[0]] += (
                    r"\qwx[" + str(target_index_list[-1] - target_index_list[0]) + r"]"
                )
                continue

            for target_index in target_index_list:
                name_latex = ""
                if len(target_index_list) == 1:
                    if name_latex_part == r"\targ":
                        name_latex = name_latex_part
                    else:
                        name_latex = r"\gate" + name_latex_part
                    gate_latex_part[target_index] = name_latex
                else:
                    rle = []
                    index = target_index_list[0]
                    size = 1
                    for t_index in target_index_list[1:]:
                        if t_index == index + size:
                            size += 1
                        else:
                            rle.append((index, size))
                            index = t_index
                            size = 1
                    rle.append((index, size))
                    for rle_index in range(len(rle)):
                        index, size = rle[rle_index]
                        if size == 1:
                            name_latex = r"\gate" + name_latex_part
                            gate_latex_part[index] = name_latex
                        else:
                            name_latex = (
                                r"\multigate{" + str(size - 1) + r"}" + name_latex_part
                            )
                            gate_latex_part[index] = name_latex
                            for k in range(1, size):
                                name_latex = r"\ghost" + name_latex_part
                                gate_latex_part[index + k] = name_latex
                        if rle_index + 1 < len(rle):
                            next_index = rle[rle_index + 1][0]
                            gate_latex_part[index + size - 1] += (
                                r"\qwx[" + str(next_index - index - size + 1) + r"]"
                            )

            for control_index in control_index_list:
                gate_latex_part[control_index] = (
                    r"\ctrl{" + str(target_index_list[0] - control_index) + r"}"
                )

        for qubit in range(qubit_count):
            gate_latex[qubit].append(gate_latex_part[qubit])

    circuit_latex_part = [" & ".join(gate_latex[i]) for i in range(qubit_count)]
    circuit_latex = (r" \\" + "\n").join(circuit_latex_part)

    res = r"""
\documentclass[border=2px]{standalone}
\usepackage[braket, qm]{qcircuit}
\usepackage{graphicx}

\begin{document}
    \scalebox{1.0}{
    \Qcircuit @C=1.0em @R=0.2em @!R { \\
"""
    res += circuit_latex
    res += r""" \\
    }}
\end{document}"""

    return res



import copy
from typing import List

from qulacs import QuantumCircuit
from typing_extensions import TypedDict

GATE_DEFAULT_WIDTH = 1.0
GATE_DEFAULT_HEIGHT = 1.5

GateData = TypedDict(
    "GateData",
    {
        "text": str,
        "width": float,
        "height": float,
        "raw_text": str,
        "target_bit": List[int],
        "control_bit": List[int],
    },
)

CircuitData = List[List[GateData]]


class CircuitParser:
    """
    Parse quantum circuit into a list of gate data.

    Parameters
    ----------
    circuit : QuantumCircuit
        Quantum circuit to be parsed.

    Attributes
    ----------
    qubit_count : int
        Number of qubits in the circuit.
    gate_info : CircuitData
        List of gate data.
    layer_width : List[float]
        Width of each layer.
    gate_dict : Dict[str, str]
        A dictionary mapping gate names to their latex representation.
    """

    def __init__(self, circuit: QuantumCircuit):
        self.qubit_count = circuit.get_qubit_count()
        self.gate_info: CircuitData = [[] for _ in range(self.qubit_count)]
        self.layer_width = []

        self.gate_dict = {
            "I": r"$I$",
            "X": r"$X$",
            "Y": r"$Y$",
            "Z": r"$Z$",
            "H": r"$H$",
            "S": r"$S$",
            "Sdag": r"$S^\dag$",
            "T": r"$T$",
            "Tdag": r"$T^\dag$",
            "sqrtX": r"$\sqrt{X}$",
            "sqrtXdag": r"$\sqrt{X^\dag}$",
            "sqrtY": r"$\sqrt{Y}$",
            "sqrtYdag": r"$\sqrt{Y^\dag}$",
            "Projection-0": r"$P0$",
            "Projection-1": r"$P1$",
            "U1": r"$U1$",
            "U2": r"$U2$",
            "U3": r"$U3$",
            "X-rotation": r"$RX$",
            "Y-rotation": r"$RY$",
            "Z-rotation": r"$RZ$",
            "Pauli": r"$Pauli$",
            "Pauli-rotation": r"$PR$",
            "CZ": r"$CZ$",
            "CNOT": r"$\targ$",
            "SWAP": r"$SWAP$",
            "Reflection": r"$Ref$",
            "ReversibleBoolean": r"$ReB$",
            "DenseMatrix": r"$DeM$",
            "DinagonalMatrix": r"$DiM$",
            "SparseMatrix": r"$SpM$",
            "Generic gate": r"$GeG$",
            "ParametricRX": r"$pRX$",
            "ParametricRY": r"$pRY$",
            "ParametricRZ": r"$pRZ$",
            "ParametricPauliRotation": r"$pPR$",
        }

        default_value: GateData = {
            "raw_text": "wire",
            "width": GATE_DEFAULT_WIDTH,
            "height": GATE_DEFAULT_HEIGHT,
            "text": "",
            "target_bit": [],
            "control_bit": [],
        }
        layer_info: List[GateData] = [
            copy.deepcopy(default_value) for _ in range(self.qubit_count)
        ]
        gate_num = circuit.get_gate_count()

        for i in range(gate_num):
            gate = circuit.get_gate(i)
            if len(gate.get_target_index_list()) == 0:
                print(
                    """CAUTION: The {}-th Gate you added is skipped.\
                    This gate does not have "target_qubit_list".""".format(
                        i
                    )
                )
                continue
            target_index_list = gate.get_target_index_list()
            control_index_list = gate.get_control_index_list()
            index_list = target_index_list + control_index_list
            gate_name = gate.get_name()
            name_latex = self.gate_dict[gate_name]

            conflict = False
            for index in range(min(index_list), max(index_list) + 1):
                if layer_info[index]["raw_text"] != "wire":
                    conflict = True
            if conflict:
                self.append_layer(layer_info, default_value)

            for index in range(min(index_list), max(index_list) + 1):
                layer_info[index]["raw_text"] = "used"
            for target_index in target_index_list:
                if target_index == target_index_list[0]:
                    layer_info[target_index]["raw_text"] = gate_name
                    layer_info[target_index]["text"] = name_latex
                    layer_info[target_index]["width"] = GATE_DEFAULT_WIDTH
                    layer_info[target_index]["height"] = GATE_DEFAULT_HEIGHT
                    layer_info[target_index]["target_bit"] = target_index_list
                    layer_info[target_index]["control_bit"] = control_index_list
                else:
                    layer_info[target_index]["width"] = GATE_DEFAULT_WIDTH
                    layer_info[target_index]["raw_text"] = "ghost"

        self.append_layer(layer_info, default_value)

        self.layer_width = [GATE_DEFAULT_WIDTH for _ in range(len(self.gate_info[0]))]
        for i in range(len(self.gate_info[0])):
            for j in range(self.qubit_count):
                self.layer_width[i] = max(
                    self.layer_width[i], self.gate_info[j][i]["width"]
                )
            for j in range(self.qubit_count):
                self.gate_info[j][i]["width"] = self.layer_width[i]

    def append_layer(self, layer_info: List[GateData], default_value: GateData) -> None:
        """
        Append a layer to the layer_info.

        Parameters
        ----------
        layer_info : List[GateData]
            A list of gate data.
        default_value : GateData
            A default value of gate data.
        """
        is_blank = True
        for qubit in range(self.qubit_count):
            if layer_info[qubit]["raw_text"] != "wire":
                is_blank = False
        if not is_blank:
            for qubit in range(self.qubit_count):
                if layer_info[qubit]["raw_text"] == "used":
                    layer_info[qubit]["raw_text"] = "wire"
            for qubit in range(self.qubit_count):
                self.gate_info[qubit].append(layer_info[qubit])
                layer_info[qubit] = copy.deepcopy(default_value)



from typing import List, Tuple

import matplotlib
from matplotlib import patches
from matplotlib import pyplot as plt
from qulacs import QuantumCircuit
from typing_extensions import Final

from .circuit_parser import (
    GATE_DEFAULT_HEIGHT,
    GATE_DEFAULT_WIDTH,
    CircuitData,
    CircuitParser,
    GateData,
)

GATE_MARGIN_RIGHT: Final[float] = 0.5
GATE_MARGIN_LEFT: Final[float] = 0.5
GATE_MARGIN_BOTTOM: Final[float] = 0.5
GATE_MARGIN_TOP: Final[float] = 0.5

PORDER_LINE: Final[int] = 2
PORDER_GATE: Final[int] = 3
PORDER_TEXT: Final[int] = 4


class MPLCircuitlDrawer:
    """
    Drawing a circuit using Matplotlib.

    Parameters
    ----------
    circuit : QuantumCircuit
        A quantum circuit to be drawn.
    dpi : int optional default=72
        The resolution of the figure.
    scale : float optional default=0.6
        The scale of the figure.

    Attributes
    ----------
    _figure : matplotlib.figure.Figure
        The figure of the circuit.
    _ax : matplotlib.axes.Axes
        The axes of the figure.
    _circuit : QuantumCircuit
        The quantum circuit to be drawn.
    _parser : CircuitParser
        The parser of the quantum circuit.
    _circuit_data : CircuitData
        The data of the quantum circuit after parsing by CircuitParser.
    _fig_scale_factor : float
        The scale factor of the figure.

    Examples
    --------
    >>> from qulacs import QuantumCircuit
    >>> from qulacsvis.visualization import MPLCircuitlDrawer
    >>> import matplotlib.pyplot as plt
    >>>
    >>> circuit = QuantumCircuit(3)
    >>> circuit.add_X_gate(0)
    >>> circuit.add_Y_gate(1)
    >>> circuit.add_Z_gate(2)
    >>>
    >>> drawer = MPLCircuitlDrawer(circuit)
    >>> drawer.draw()
    >>> plt.show()
    """

    def __init__(self, circuit: QuantumCircuit, *, dpi: int = 72, scale: float = 0.6):
        self._figure = plt.figure(dpi=dpi)
        self._figure.subplots_adjust(left=0, right=1, bottom=0, top=1)
        self._ax = self._figure.add_subplot(111)
        self._ax.set_aspect("equal")
        self._ax.axis("off")

        self._circuit = circuit
        self._parser = CircuitParser(circuit)
        self._circuit_data: CircuitData = self._parser.gate_info
        self._fig_scale_factor = scale

    def draw(self, *, debug: bool = False) -> matplotlib.figure.Figure:
        """
        Draw the circuit.

        Parameters
        ----------
        debug : bool optional default=False
            If True, draw the circuit with the axes of the figure.

        Returns
        -------
        self._figure : matplotlib.figure.Figure
            The figure of the circuit.
        """
        if debug:
            self._ax.axis("on")
            self._ax.grid()

        circuit_layer_count = len(self._circuit_data[0])
        # When the input is an empty quantum circuit
        if self._parser.layer_width == []:
            self._parser.layer_width = [GATE_DEFAULT_WIDTH]
        # X/Y coordinates of the area where the circuit will be drawn.
        # Used to resize the figure.
        # In particular, the X coordinate is also used
        # as the right end (max_x) and left end (min_x) coordinates of the circuit wire.
        circuit_max_x = (
            sum(self._parser.layer_width)
            + circuit_layer_count * GATE_MARGIN_RIGHT
            - self._parser.layer_width[0] / 2
        )
        circuit_min_x = -self._parser.layer_width[0] / 2 - GATE_MARGIN_LEFT
        circuit_max_y = (
            self._parser.qubit_count * (GATE_DEFAULT_HEIGHT + GATE_MARGIN_BOTTOM)
            - GATE_MARGIN_BOTTOM
            - GATE_DEFAULT_HEIGHT / 2
        )

        # Determine the size of the figure (drawing size of the circuit + margins)
        QUBIT_LABEL_WIDTH = 2
        self._ax.set_xlim(
            circuit_min_x - QUBIT_LABEL_WIDTH, circuit_max_x + GATE_MARGIN_RIGHT
        )
        self._ax.set_ylim(
            circuit_max_y + GATE_MARGIN_BOTTOM,
            -GATE_DEFAULT_HEIGHT / 2 - GATE_MARGIN_TOP,
        )

        # Enlarge/reduce the shape while keeping the aspect ratio
        fig_width = abs(self._ax.get_xlim()[1] - self._ax.get_xlim()[0])
        fig_heigth = abs(self._ax.get_ylim()[1] - self._ax.get_ylim()[0])
        self._figure.set_size_inches(
            fig_width * self._fig_scale_factor, fig_heigth * self._fig_scale_factor
        )

        # Draw a Qubit label for the number of Qubits in the quantum circuit and a wire
        for qubit in range(self._parser.qubit_count):
            line_ypos = qubit * (GATE_DEFAULT_HEIGHT + GATE_MARGIN_BOTTOM)
            self._text(
                circuit_min_x - 1,
                line_ypos,
                r"$q_{" + str(qubit) + r"}$",
                fontsize=30,
            )

            self._line(
                (circuit_min_x, line_ypos),
                (
                    circuit_max_x,
                    line_ypos,
                ),
            )

        # x-coordinate of the layer currently being drawn
        layer_xpos = 0.0

        # Draw a gate for each layer
        for layer in range(circuit_layer_count):
            for qubit in range(self._parser.qubit_count):
                gate = self._circuit_data[qubit][layer]
                qubit_ypos = qubit * (GATE_DEFAULT_HEIGHT + GATE_MARGIN_BOTTOM)
                if gate["raw_text"] == "ghost":
                    continue
                elif gate["raw_text"] == "wire":
                    continue
                elif gate["raw_text"] == "CNOT":
                    self._cnot(gate, (layer_xpos, qubit_ypos))
                elif gate["raw_text"] == "SWAP":
                    self._swap(gate, (layer_xpos, qubit_ypos))
                elif len(gate["target_bit"]) > 1:
                    self._multi_gate(gate, (layer_xpos, qubit_ypos))
                else:
                    self._gate_with_size(gate, (layer_xpos, qubit_ypos), 1)

            # Determine the x-coordinate of the next layer
            layer_xpos += self._parser.layer_width[layer] + GATE_MARGIN_RIGHT

        return self._figure

    def _line(
        self,
        from_xy: Tuple[float, float],
        to_xy: Tuple[float, float],
        lc: str = "k",
        ls: str = "-",
        lw: float = 2.0,
        zorder: int = PORDER_LINE,
    ) -> None:
        """
        Draw a line.

        Parameters
        ----------
        from_xy : Tuple[float, float]
            The position of the start point.
        to_xy : Tuple[float, float]
            The position of the end point.
        lc : str optional default="k"
            The color of the line.
        ls : str optional default="-"
            The style of the line.
        lw : float optional default=2.0
            The width of the line.
        zorder : int optional default=PORDER_LINE
            The zorder of the line.
        """

        from_x, from_y = from_xy
        to_x, to_y = to_xy
        self._ax.plot(
            [from_x, to_x],
            [from_y, to_y],
            color=lc,
            linestyle=ls,
            linewidth=lw,
            zorder=zorder,
        )

    def _text(
        self,
        x: float,
        y: float,
        text: str,
        horizontalalignment: str = "center",
        verticalalignment: str = "center",
        fontsize: int = 20,
        color: str = "k",
        clip_on: bool = True,
        zorder: int = PORDER_TEXT,
    ) -> None:
        """
        Draw a text.

        Parameters
        ----------
        x : float
            The x position of the text.
        y : float
            The y position of the text.
        text : str
            The text to be drawn.
        horizontalalignment : str optional default="center"
            The horizontal alignment of the text.
        verticalalignment : str optional default="center"
            The vertical alignment of the text.
        fontsize : int optional default=20
            The font size of the text.
        color : str optional default="k"
            The color of the text.
        clip_on : bool optional default=True
            If True, the text will be clipped.
        zorder : int optional default=PORDER_TEXT
            The zorder of the text.

        Notes
        -----
        The fontsize will be adjusted by _fig_scale_factor.
        """

        self._ax.text(
            x,
            y,
            text,
            horizontalalignment=horizontalalignment,
            verticalalignment=verticalalignment,
            fontsize=fontsize * self._fig_scale_factor,
            color=color,
            clip_on=clip_on,
            zorder=zorder,
        )

    def _gate_with_size(
        self, gate: GateData, xy: Tuple[float, float], multi_gate_size: int
    ) -> None:
        """
        Draw a gate with a specified size.

        Parameters
        ----------
        gate : GateData
            The gate data to be drawn.
        xy : Tuple[float, float]
            The position of the gate with the smallest index among the target bits.
        multi_gate_size : int
            The size of the gate.
        """

        xpos, ypos = xy
        # Find the midpoint of the y-coordinate of the gate with size
        ypos = (
            ypos
            + (
                ypos
                + (multi_gate_size - 1) * (GATE_DEFAULT_HEIGHT + GATE_MARGIN_BOTTOM)
            )
        ) * 0.5
        multi_gate_height = (
            GATE_DEFAULT_HEIGHT * multi_gate_size
            + GATE_MARGIN_BOTTOM * (multi_gate_size - 1)
        )
        box = patches.Rectangle(
            # The gate is centered.
            xy=(xpos - 0.5 * gate["width"], ypos - 0.5 * multi_gate_height),
            width=gate["width"],
            height=multi_gate_height,
            facecolor="w",
            edgecolor="k",
            linewidth=2.4,
            zorder=PORDER_GATE,
        )
        self._ax.add_patch(box)
        self._text(xpos, ypos, gate["text"])
        self._control_bits(gate["control_bit"], (xpos, ypos))

    def _multi_gate(self, gate: GateData, xy: Tuple[float, float]) -> None:
        """
        Draw a multi-gate. (e.g., DensityMatrixGate)

        Parameters
        ----------
        gate : GateData
            The gate data to be drawn.
        xy : Tuple[float, float]
            The position of the gate with the smallest index among the target bits.
        """

        xpos, ypos = xy

        multi_gate_data: GateData = {
            "text": gate["text"],
            "width": gate["width"],
            "height": gate["height"],
            "target_bit": [],
            "control_bit": [],
            "raw_text": gate["raw_text"],
        }

        groups_of_adjacent_gates = self._grouping_adjacent_gates(gate["target_bit"])

        for i, adjacent_gates in enumerate(groups_of_adjacent_gates):
            group_x = xpos
            group_y = adjacent_gates[0] * (GATE_DEFAULT_HEIGHT + GATE_MARGIN_BOTTOM)
            self._gate_with_size(
                multi_gate_data, (group_x, group_y), len(adjacent_gates)
            )
            if i == 0:
                # Show the name only for the first group (multi_gate) and hide the rest
                multi_gate_data["text"] = ""

        # Gray line connecting the gates
        ypos = min(gate["target_bit"]) * (GATE_DEFAULT_HEIGHT + GATE_MARGIN_BOTTOM)
        to_ypos = max(gate["target_bit"]) * (GATE_DEFAULT_HEIGHT + GATE_MARGIN_BOTTOM)
        self._line((xpos, ypos), (xpos, to_ypos), lw=10, lc="gray")

        self._control_bits(gate["control_bit"], (xpos, ypos))

    def _grouping_adjacent_gates(self, target_bits: List[int]) -> List[List[int]]:
        """
        Grouping adjacent gates.

        Parameters
        ----------
        target_bit : List[int]
            The target bit list.

        Returns
        -------
        List[List[int]]
            The grouped target bit list.

        Examples
        --------
        >>> target_bits = [1, 2, 3, 5, 7, 8]
        >>> print(_grouping_adjacent_gates(target_bits))
        >>> [[1, 2, 3], [5], [7, 8]]
        """

        target_bits.sort()
        groups = []
        adjacent_gates: List[int] = []
        for target_bit in target_bits:
            if adjacent_gates == []:
                adjacent_gates.append(target_bit)
                continue

            if target_bit - 1 in adjacent_gates:
                adjacent_gates.append(target_bit)
            else:
                groups.append(adjacent_gates)
                adjacent_gates = [target_bit]

        groups.append(adjacent_gates)
        return groups

    def _cnot(self, gate: GateData, xy: Tuple[float, float]) -> None:
        """
        Draw a CNOT gate.

        Parameters
        ----------
        gate : GateData
            The gate data to be drawn.
        xy : Tuple[float, float]
            The position of the gate indicating the target bit of CNOT.

        Raise
        -----
        ValueError
            If the gate does not have a control bit.
        """

        xpos, ypos = xy
        TARGET_QUBIT_RADIUS: Final[float] = 0.4

        if gate["control_bit"] is None:
            raise ValueError("control_bit is None")
        elif gate["control_bit"] == []:
            raise ValueError("control_bit is empty")

        target = patches.Circle(
            xy=(xpos, ypos),
            radius=TARGET_QUBIT_RADIUS,
            fc="w",
            ec="k",
            zorder=PORDER_GATE,
        )
        self._ax.add_patch(target)
        # draw target qubit's "+" mark
        # |
        self._line(
            (xpos, ypos - TARGET_QUBIT_RADIUS),
            (xpos, ypos + TARGET_QUBIT_RADIUS),
            zorder=PORDER_TEXT,
        )
        # -
        self._line(
            (xpos - TARGET_QUBIT_RADIUS, ypos),
            (xpos + TARGET_QUBIT_RADIUS, ypos),
            zorder=PORDER_TEXT,
        )

        self._control_bits(gate["control_bit"], (xpos, ypos))

    def _control_bits(
        self, control_bits: List[int], xy_from: Tuple[float, float]
    ) -> None:
        """
        Draw control bits.

        Parameters
        ----------
        control_bits : List[int]
            The control bits to be drawn.
        xy_from : Tuple[float, float]
            The position of the gate from which the control bits are connected.
        """

        for control_bit in control_bits:
            to_ypos = control_bit * (GATE_DEFAULT_HEIGHT + GATE_MARGIN_RIGHT)
            to_xpos = xy_from[0]

            self._line(
                xy_from,
                (to_xpos, to_ypos),
            )
            ctl = patches.Circle(
                xy=(to_xpos, to_ypos),
                radius=0.15,
                fc="k",
                ec="w",
                linewidth=0,
                zorder=PORDER_GATE,
            )
            self._ax.add_patch(ctl)

    def _swap(self, gate: GateData, xy: Tuple[float, float]) -> None:
        """
        Draw a SWAP gate.

        Parameters
        ----------
        gate : GateData
            The gate data to be drawn.
        xy : Tuple[float, float]
            The position of the gate with the smaller index among the target bits.
        """

        xpos, ypos = xy
        TARGET_QUBIT_MARK_SIZE: Final[float] = 0.2

        for target_bit in gate["target_bit"]:
            to_ypos = target_bit * (GATE_DEFAULT_HEIGHT + GATE_MARGIN_RIGHT)
            self._line(
                (xpos, ypos),
                (xpos, to_ypos),
            )
            # draw target qubit's "x" mark
            # \
            self._line(
                (xpos - TARGET_QUBIT_MARK_SIZE, to_ypos - TARGET_QUBIT_MARK_SIZE),
                (xpos + TARGET_QUBIT_MARK_SIZE, to_ypos + TARGET_QUBIT_MARK_SIZE),
            )
            # /
            self._line(
                (xpos + TARGET_QUBIT_MARK_SIZE, to_ypos - TARGET_QUBIT_MARK_SIZE),
                (xpos - TARGET_QUBIT_MARK_SIZE, to_ypos + TARGET_QUBIT_MARK_SIZE),
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qulacsvis', 'qulacsvis.utils', 'qulacsvis.visualization']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.3.2,<9.0.0',
 'Qulacs>=0.3.0,<0.4.0',
 'matplotlib>=3.4.3,<4.0.0',
 'numpy>=1.21.2,<2.0.0',
 'temp>=2020.7.2,<2021.0.0']

setup_kwargs = {
    'name': 'qulacsvis',
    'version': '0.2.1',
    'description': 'visualizers for qulacs',
    'long_description': '# qulacs-visualizer\n\n[![CI](https://github.com/Qulacs-Osaka/qulacs-visualizer/actions/workflows/ci.yml/badge.svg)](https://github.com/Qulacs-Osaka/qulacs-visualizer/actions/workflows/ci.yml) [![Build and Deploy Documentation](https://github.com/Qulacs-Osaka/qulacs-visualizer/actions/workflows/doc.yml/badge.svg)](https://github.com/Qulacs-Osaka/qulacs-visualizer/actions/workflows/doc.yml) [![PyPI version](https://badge.fury.io/py/qulacsvis.svg)](https://badge.fury.io/py/qulacsvis) [![MIT](https://img.shields.io/badge/license-MIT-blue.svg?style=flat)](LICENSE)\n\nqulacs-visualizer is a quantum circuit drawing library for [qulacs](https://github.com/qulacs/qulacs). This library only supports Python. Not available in C/C++.\n\nqulacs-visualizer supports the following methods.\n\n- Text-Based Drawing\n- Matplotlib Drawing\n  - using [matplotlib](https://github.com/matplotlib/matplotlib)\n- LaTeX Drawing\n  - using LaTeX and the [qcircuit package](https://github.com/CQuIC/qcircuit)\n\n\n## Quick Install\n\nInstall using `pip` from PyPI:\n\n```\npip install qulacsvis\n```\n\n## Example\n\n\nFor more information, see the [documentation](https://qulacs-osaka.github.io/qulacs-visualizer/qulacsvis.visualization.circuit_drawer.html) of the `circuit_drawer()` function.\nThe documentation describes the other options available.\n\n### Text-Based Drawing\n\nIn text-based mode, draws the circuit as ASCII art. This mode is the default behavior.\n\n```py\nfrom qulacs import QuantumCircuit\nfrom qulacsvis import circuit_drawer\n\n# Build a quantum circuit\ncircuit = QuantumCircuit(3)\ncircuit.add_X_gate(0)\ncircuit.add_Y_gate(1)\ncircuit.add_Z_gate(2)\ncircuit.add_dense_matrix_gate(\n    [0, 1], [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]\n)\ncircuit.add_CNOT_gate(2, 0)\ncircuit.add_X_gate(2)\n\n# Draw a quantum circuit\ncircuit_drawer(circuit)\n```\n```\n   ___     ___     ___\n  | X |   |DeM|   |CX |\n--|   |---|   |---|   |----------\n  |___|   |   |   |___|\n   ___    |   |     |\n  | Y |   |   |     |\n--|   |---|   |-----|------------\n  |___|   |___|     |\n   ___              |      ___\n  | Z |             |     | X |\n--|   |-------------●-----|   |--\n  |___|                   |___|\n```\n\n### Matplotlib Drawing\n\nTo use another drawing method, you can specify it by setting a value to the `output_method` argument of the `circuit_drawer()` function. For matplotlib drawing, set `output_method="mpl"`.\n\n```py\ncircuit_drawer(circuit, "mpl")\n```\n\n![circuit_matplotlib_drawing.png](doc/source/_static/circuit_matplotlib_drawing.png)\n\n## LaTeX Drawing\n\nFor LaTeX drawing, set `output_method="latex"`.\n\n```py\ncircuit_drawer(circuit, "latex")\n```\n\n![circuit_latex_drawing.png](doc/source/_static/circuit_latex_drawing.png)\n\nIf you want to output LaTeX code, set `output_method="latex_source"`.\n\n```py\nprint(circuit_drawer(circuit, "latex_source"))\n```\n\n```latex\n\\documentclass[border=2px]{standalone}\n\\usepackage[braket, qm]{qcircuit}\n\\usepackage{graphicx}\n\n\\begin{document}\n    \\scalebox{1.0}{\n    \\Qcircuit @C=1.0em @R=0.2em @!R { \\\\\n        \\nghost{ {q}_{0} : } & \\lstick{ {q}_{0} :  } & \\gate{X} & \\multigate{1}{DeM} & \\targ & \\qw \\\\\n        \\nghost{ {q}_{1} : } & \\lstick{ {q}_{1} :  } & \\gate{Y} & \\ghost{DeM} & \\qw & \\qw \\\\\n        \\nghost{ {q}_{2} : } & \\lstick{ {q}_{2} :  } & \\gate{Z} & \\qw & \\ctrl{-2} & \\gate{X} \\\\\n    }}\n\\end{document}\n```\n\n### Requirement\n\nIf you want to use LaTeX Drawing, you need to have a local environment where you can run LaTeX (pdflatex).\nYou will also need the [qcircuit package](https://github.com/CQuIC/qcircuit).\n[TeX Live](https://www.tug.org/texlive/) and [MiKTeX](https://miktex.org/) have the qcircuit package installed by default.\n\n## License\n\n[MIT License](LICENSE)\n\n## Contributors\n\nWe use [qqcd](https://github.com/mf-22/qqcd) for text-based drawing, a drawing library developed by [@mf-22](https://github.com/mf-22). Thank you.\n',
    'author': 'Qulacs-Osaka',
    'author_email': 'you@example.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Qulacs-Osaka/qulacs-visualizer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
